
###########################################################################
# R code for running the multi-level intercept-only model with JAGS/rjags #
###########################################################################

# Supplementary material to article:
# Hecht, M., Weirich, S., & Zitzmann, S. (2021). Comparing the MCMC efficiency
#   of JAGS and Stan for the multi-level intercept-only model in the covariance-
#   and mean-based and classic parametrization. Psych.

## used R version: 4.0.5 (R Core Team, 2021)
# => get latest version: https://cran.r-project.org/

## install JAGS from
# https://sourceforge.net/projects/mcmc-jags/files
# used JAGS version: 4.3.0 (Plummer, 2017)

## load/install packages
# used rjags version: 4-10 (Plummer, 2019)
# => get latest rjags version: install.packages( "rjags" )
library( rjags )
# used shinystan version: 2.5.0 (Gabry, 2018)
# => get latest shinystan version: install.packages( "shinystan" )
library( shinystan )

## bugs files
# covariance-/mean-based parametrization
bugs.file <- c("https://figshare.com/ndownloader/files/31595090")
# for classic parametrization use:
# bugs.file <- c("https://figshare.com/ndownloader/files/31595099")

## open bugs file
bugs.file <- url( bugs.file )

## load simulated example data
(load(url("https://figshare.com/ndownloader/files/31595084")))
# D: data matrix in wide format:
#    rows:     level 1 units (e.g., measurement occasions)
#    columns:  level 2 units (e.g., persons)
#    cells:    values
# J: number of level 2 units
# P: number of level 1 units


#####################
### preprocessing ###
#####################

## compute sample mean vector and sample scatter matrix from data
# sample mean vector y.bar
y.bar <- matrix( rowMeans( D ), P, 1 )
# column vector of ones
ones.vec <- matrix( 1, J, 1 )
# sample scatter matrix
S <- ( D - y.bar %*% t(ones.vec) ) %*% t( ( D - y.bar %*% t(ones.vec) ) )

## data list for JAGS
data.list <- list()
data.list <- c( data.list, list( "S"=S ) )
data.list <- c( data.list, list( "y.bar"=y.bar ) )
data.list <- c( data.list, list( "J"=J ) )
data.list <- c( data.list, list( "P"=P ) )
data.list <- c( data.list, list( "D"=D ) ) # only needed for classic parametrization

## function to generate random starting values
# (starting values are true values from data generation: mu = 0, var.theta = 1, var.eps = 1)
generate.startval <- function(){
   startval <- list()
   startval <- c( startval, list( "mu"         = 0   ) )
   startval <- c( startval, list( "prec.theta" = 1/1 ) )
   startval <- c( startval, list( "prec.eps"   = 1/1 ) )
   return( startval )
}
# optional for classic parametrization:
# add starting values for parameters of level 2 units (theta)

# set starting values for the parameters (one chain each)
inits <- sapply( 1:1, function (chain.nr,l) l, generate.startval(), simplify=FALSE )

# set seeds for the chain
inits[[1]]$".RNG.name" <- "base::Wichmann-Hill"
inits[[1]]$".RNG.state" <- c( 11076, 3733, 19174 )


##############
### warmup ###
##############

# warmup iterations
warmup.iterations <- 1000

# warmup start time
start.time.warmup <- Sys.time()

# set up & warm up JAGS model
ini <- jags.model( file=bugs.file, data=data.list,
                   n.chains=1, n.adapt=warmup.iterations,
                   inits=inits, quiet=TRUE )

# warmup runtime
( runtime.warmup <- Sys.time() - start.time.warmup )

# close bugs file
close( bugs.file )


################
### sampling ###
################

# sampling iterations
sampling.iterations <- 100000

# sampling start time
start.time.sampling <- Sys.time()

# sampling
res <- jags.samples( ini, variable.names=c("mu","sigma2.theta","sigma2.eps"),
                     n.iter=sampling.iterations, thin=1, type='trace', progress.bar = NULL )
# optional for classic parametrization:
# add "theta" to argument variable.names

# sampling runtime
( runtime.sampling <- Sys.time() - start.time.sampling )

# set time units of sampling runtime to seconds (needed below to compute ESS/s)
units( runtime.sampling ) <- "secs"


######################
### postprocessing ###
######################

# create shinystan object
samples <- cbind( matrix( res$mu[1,,1], ncol=1 ),
                  matrix( res$sigma2.theta[1,,1], ncol=1 ),
                  matrix( res$sigma2.eps[1,,1], ncol=1 ) )
colnames( samples ) <- c( "mu", "sigma2.theta", "sigma2.eps" )
shinystan.obj <- as.shinystan( list( samples ) )

# retrieve effective sample size (ESS)
( ESS <- retrieve( shinystan.obj, "ess" ) )

# calculate MCMC efficiency (ESS/s)
( MCMC.efficiency <- ESS / as.numeric( runtime.sampling ) )
