
###########################################################################
# R code for running the multi-level intercept-only model with Stan/rstan #
###########################################################################

# Supplementary material to article:
# Hecht, M., Weirich, S., & Zitzmann, S. (2021). Comparing the MCMC efficiency
#   of JAGS and Stan for the multi-level intercept-only model in the covariance-
#   and mean-based and classic parametrization. Psych.

## used R version: 4.0.5 (R Core Team, 2021)
# => get latest version: https://cran.r-project.org/

## load/install packages
# used rstan version: 2.21.2 (Stan Development Team, 2020)
# => get latest rstan version: install.packages( "rstan" )
library( rstan )
# used shinystan version: 2.5.0 (Gabry, 2018)
# => get latest shinystan version: install.packages( "shinystan" )
library( shinystan )

## working directory
# (!!! needs to include file 02b_multilevel_intonly_model_covmeanbased.stan !!!)
wd <- "SET_YOUR_WORKING_DIRECTORY_HERE___needs_to_include_file___02b_multilevel_intonly_model_covmeanbased.stan"
# for classic parametrization include: 02c_multilevel_intonly_model_classic.stan

## load simulated example data
(load(url("https://figshare.com/ndownloader/files/31595084")))
# D: data matrix in wide format:
#    rows:     level 1 units (e.g., measurement occasions)
#    columns:  level 2 units (e.g., persons)
#    cells:    values
# J: number of level 2 units
# P: number of level 1 units


#####################
### preprocessing ###
#####################

## compute sample mean vector and sample scatter matrix from data
# sample mean vector y.bar
y.bar <- matrix( rowMeans( D ), P, 1 )
# column vector of ones
ones.vec <- matrix( 1, J, 1 )
# sample scatter matrix
S <- ( D - y.bar %*% t(ones.vec) ) %*% t( ( D - y.bar %*% t(ones.vec) ) )

## data list for Stan
data.list <- list()
data.list <- c( data.list, list( "S"=S ) )
data.list <- c( data.list, list( "ybar"=as.vector( y.bar ) ) )
data.list <- c( data.list, list( "J"=J ) )
data.list <- c( data.list, list( "P"=P ) )
data.list <- c( data.list, list( "D"=D ) ) # only needed for classic parametrization


#########################
### model translation ###
#########################

# translation start time
start.time.translation <- Sys.time()

# translate model
translated <- stanc( file = file.path( wd, "02b_multilevel_intonly_model_covmeanbased.stan" ), isystem = wd )
# for classic parametrization use: 02c_multilevel_intonly_model_classic.stan
# translated <- stanc( file = file.path( wd, "02c_multilevel_intonly_model_classic.stan" ), isystem = wd )

# translation runtime
( runtime.translation <- Sys.time() - start.time.translation )


#########################
### model compilation ###
#########################

# compilation start time
start.time.compilation <- Sys.time()

# compile model
compiled  <- stan_model( stanc_ret = translated, auto_write = TRUE, isystem = wd )

# compilation runtime
( runtime.compilation <- Sys.time() - start.time.compilation )


################################################
### warmup/sampling with function sampling() ###
################################################

# samplingfunction start time
start.time.samplingfunction <- Sys.time()

# warmup/sampling iterations
warmup.iterations <- 1000
sampling.iterations <- 100000

# warmup/sampling
res <- sampling( object = compiled, data = data.list, chains = 1, cores = 1,
                 iter = warmup.iterations+sampling.iterations, warmup = warmup.iterations,
                 init = list( list( mu = 0, sigma2theta = 1, sigma2eps = 1 ) ),
                 seed = 12345 )
# optional for classic parametrization:
# add starting values (argument init) for parameters of level 2 units (theta)

# samplingfunction runtime
( runtime.samplingfunction <- Sys.time() - start.time.samplingfunction )


######################
### postprocessing ###
######################

# times reported by rstan
runtimes <- attributes( res@sim$samples[[1]] )[["elapsed_time"]]

# time for warmup
( runtime.warmup <- runtimes[["warmup"]] )

# time for sampling
( runtime.sampling <- runtimes[["sample"]] )

# create shinystan object
samples <- cbind( matrix( res@sim[["samples"]][[1]]$mu[(warmup.iterations+1):(warmup.iterations+sampling.iterations)], ncol=1 ),
                  matrix( res@sim[["samples"]][[1]]$sigma2theta[(warmup.iterations+1):(warmup.iterations+sampling.iterations)], ncol=1 ),
                  matrix( res@sim[["samples"]][[1]]$sigma2eps[(warmup.iterations+1):(warmup.iterations+sampling.iterations)], ncol=1 ) )
colnames( samples ) <- c( "mu", "sigma2.theta", "sigma2.eps" )
shinystan.obj <- as.shinystan( list( samples ) )

# retrieve effective sample size (ESS)
( ESS <- retrieve( shinystan.obj, "ess" ) )

# calculate MCMC efficiency (ESS/s)
( MCMC.efficiency <- ESS / as.numeric( runtime.sampling ) )
