// Stan model code for the multi-level intercept-only model
// in the covariance- and mean-based parametrization

// Supplementary material to article:
// Hecht, M., Weirich, S., & Zitzmann, S. (2021). Comparing the MCMC efficiency
//   of JAGS and Stan for the multi-level intercept-only model in the covariance-
//   and mean-based and classic parametrization. Psych.

data {
   matrix[20,20] S; // sample scatter matrix
   vector[20] ybar; // sample mean vector
   int<lower=1> J;  // number of level 2 units
   int<lower=1> P;  // number of level 1 units
}

parameters {
   real<lower=0> sigma2theta;
   real<lower=0> sigma2eps;
   real mu;
}

transformed parameters {

   matrix[20,20] Sigma;
   matrix[20,20] SigmaM;
   vector[20] muvec;
   
   // construction of covariance matrix Sigma (Equation 5)
   // main diagonal
   for ( p in 1:P ){
      Sigma[p,p] = sigma2theta + sigma2eps;
   }
   // lower/upper triangle
   for (k in 2:P){
        for (l in 1:(k-1)) {
             Sigma[k,l] = sigma2theta;
             Sigma[l,k] = Sigma[k,l];
        }
   }
   
   // covariance matrix of means
   SigmaM = Sigma/J;
   
   // construction of mu vector
   for ( p in 1:P ){
      muvec[p] = mu;
   }
}

model {

   // distributional assumption of sample scatter matrix (Equation 3)
   S ~ wishart( J-1, Sigma );
   
   // distributional assumption of sample means (Equation 4)
   ybar ~ multi_normal( muvec, SigmaM );
   
   // prior distribution for mu
   mu ~ normal( 0, sqrt(10000) );
   
   // prior distribution for sigma2theta
   sigma2theta ~ inv_gamma(0.001,0.001);
   
   // prior distribution for sigma2eps   
   sigma2eps ~ inv_gamma(0.001,0.001);
}
