
#####################################
# R code for running the simulation #
#####################################

# Supplementary material to article:
# Hecht, M., Weirich, S., & Zitzmann, S. (2021). Comparing the MCMC efficiency
#   of JAGS and Stan for the multi-level intercept-only model in the covariance-
#   and mean-based and classic parametrization. Psych.

## used R version: 4.0.5 (R Core Team, 2021)
# => get latest version: https://cran.r-project.org/

# used car version: 3.0-10
# => get latest car version: install.packages( "car" )
require(car)

# used rjags version: 4-10 (Plummer, 2019)
# => get latest rjags version: install.packages( "rjags" )
library(rjags)

# used rstan version: 2.21.2 (Stan Development Team, 2020)
# => get latest rstan version: install.packages( "rstan" )
library(rstan)

### function to generate data from the NULL model
##  v1 21.05.2021
#   mu.true          ... population mean
#   sigma2theta.true ... between person variance
#   sigma2eps.true   ... within person varianz
#   J                ... number of level 2 units (persons)
#   P                ... number of level 1 units (observations/time points)
#   value:               list of length 2; "y": P x J wide format matrix; "seed": seed
nullmodel_data_generation <- function( mu.true = 0, sigma2theta.true = 1, sigma2eps.true = 1, J = 1000, P = 20, seed=NULL ){

		# seed
		if( is.null( seed ) ) seed <- sample( 1:99999999, 1 )
		set.seed( seed )

		# data generation
		thetaj <- rnorm( J, mu.true, sqrt(sigma2theta.true) )
		y <- sapply( thetaj, function( mu ) rnorm( P, mu, sqrt(sigma2eps.true) ), simplify=TRUE )

		# return
		list( "y"=y, "seed"=seed )
}

### simulation conditions
ss   <- 1000                                                                    ### sample size
nwarm<- c(150, 1000)                                                            ### number of warmups
prior<- c(0.001, 0.01, 0.1, 1)                                                  ### priors
iter <- 100000                                                                  ### number of iterations
reps <- 1:1000                                                                  ### number of replications
res  <- expand.grid ( nWarmups = nwarm, prior = prior, iter = iter, replication = reps, sampleSize = ss, stringsAsFactors = FALSE)
res  <- data.frame ( res, sequential_number = 1:nrow(res), stringsAsFactors = FALSE)

### define paths, need approx. 45 GB free space on hard disc
dir    <- "c:/stan_simulation"
outdir <- file.path(dir, "results")
workdir<- file.path(dir, "work")

### create folders if necessary
if(!dir.exists(outdir)) {dir.create(outdir)}
if(!dir.exists(workdir)) {dir.create(workdir)}

### jags and stan default model
jags   <- scan(url("https://figshare.com/ndownloader/files/31595090"), what="character",sep="\n",quiet=TRUE)
stan   <- scan(url("https://figshare.com/ndownloader/files/31595108"), what="character",sep="\n",quiet=TRUE)

### R functions to start jags/stan functions from R
# S ... sample scatter matrix
# y.bar ... sample mean column vector
# J ... number of level 2 units (persons)
# P ... number of level 1 units (observations/time points)
# n.iter ... number of iterations, default: 500
# warmup ... number of warmup iterations ("n.adapt"), default: 0
# bugs.file ... bugs/jags model syntax as plain text file
# inits ... list (chains) of lists (parameters) with init values;  "auto" (default) sets inits to mu=0, prec.theta=1, prec.eps=1
# time.units ... character vector, time units, e.g. "secs" (default)
# value: list; first entry "res": data.frame (1st column "parameter", further columns: iterations)
#              second entry: named vector with run times for warmup, run, total
#              third entry "seeds": generated seed for running jags
nullmodel_jags <- function( S, y.bar, J, P, n.iter=500, warmup=0, bugs.file, inits="auto", time.units="secs" ){
		## data list for JAGS
		data.list <- list()
		data.list <- c( data.list, list( "S"=S ) )
		data.list <- c( data.list, list( "y.bar"=y.bar ) )
		data.list <- c( data.list, list( "J"=J ) )
		data.list <- c( data.list, list( "P"=P ) )
		## JAGS run parameters
		# thinning
		thin <- 1
		# starting values
		if ( inits %in% "auto" ){
				# function to generate random starting values
				generate.startval <- function(){
				   startval <- list()
				   # startval <- c( startval, list( "mu"         = rnorm(1,0,sqrt(10000)) ) )
				   startval <- c( startval, list( "mu"         = 0 ) )
				   # startval <- c( startval, list( "prec.theta" = rgamma(1,0.001,0.001)+10^(-20) ) )
				   startval <- c( startval, list( "prec.theta" = 1/1 ) )
				   # startval <- c( startval, list( "prec.eps"   = rgamma(1,0.001,0.001)+10^(-20) ) )
				   startval <- c( startval, list( "prec.eps"   = 1/1 ) )
				   return( startval )
				}
				# generate starting values for the chain
				# set.seed( 1234 )
				inits <- sapply( 1:1, function (chain.nr,l) l, generate.startval(), simplify=FALSE )
		}
		# set seeds for the chain
		seeds <- parallel.seeds( 'base::BaseRNG', 1 )
		inits <- mapply ( function(i,s) c(i,s), inits, seeds, SIMPLIFY=FALSE )
	    ## initialize JAGS model
	    # number of chains = 1 (code runs only for one chain)
	    # number of adaptation iterations = 0 (no pre-sampling, just initialization)
	    start.warmup <- Sys.time()
		ini <- jags.model ( file=bugs.file, data=data.list, n.chains=1, n.adapt=warmup,
		 				   inits=inits, quiet=TRUE )
	    # run time warmup
		runtime.warmup <- Sys.time() - start.warmup
		units( runtime.warmup )<- time.units
		runtime.warmup <- as.numeric( runtime.warmup )
	    # run JAGS model
		track <- c("mu","sigma2.theta","sigma2.eps")
		start.run <- Sys.time()
		res <- jags.samples ( ini, variable.names=track, n.iter=n.iter, thin=thin,
							 type='trace' , progress.bar = NULL )
	    # run time main run
		runtime.run <- Sys.time() - start.run
		units( runtime.run )<- time.units
		runtime.run <- as.numeric( runtime.run )
		# total run time
		runtime <- runtime.warmup + runtime.run
		# named vector with runtimes
		runtimes <- c( runtime.warmup, runtime.run, runtime )
		names( runtimes ) <- c( "runtime.warmup", "runtime.run", "runtime" )
		# results
		dfr <- cbind( data.frame( "parameter"=names(res) ), do.call( "rbind", sapply( res, function(x) x[1,,1], simplify=FALSE ) ) )
		dfr <- dfr[ track, ]
		rownames( dfr ) <- seq( along=rownames( dfr ) )
		# return
		list( "samples"=dfr, "runtimes"=runtimes, "seed"=seeds )
}

### start simulation
res  <- by(data = res, INDICES = res[,"sequential_number"], FUN = function ( bed) {
        dat   <- nullmodel_data_generation(J=bed[["sampleSize"]])
        d     <- dat[["y"]]
        y.bar <- matrix( rowMeans( d ), 20, 1 )                                 ### mean vector (P = number of time points; D = matrix, persons x time points)
        ones  <- matrix( 1, ncol(d), 1 )                                        ### vector with 1 values
        S     <- ( d - y.bar %*% t(ones) ) %*% t( ( d - y.bar %*% t(ones) ) )   ### S matrix
     # first: jags. adapt priors in bugs file
        if(bed[,"prior"] != 0.001) {
           rows <- grep("dgamma", jags)
           jags[rows] <- gsub("0.001", bed[,"prior"], jags[rows])
           rows2<- grep("dnorm", jags)
           jags[rows2] <- gsub("10000", car::recode(bed[,"prior"],"0.01 = 2500; 0.1 = 100; 1 = 25"), jags[rows2])
        }
        dir.create(file.path(workdir, paste0(bed[,"sequential_number"], "_jags")))
        write(jags,file.path(workdir, paste0(bed[,"sequential_number"], "_jags"), "01b_multilevel_intonly_model.bugs"), sep="\n")
        fit   <- nullmodel_jags ( S=S, y.bar=y.bar, J=ncol(d), P=nrow(d), n.iter=bed[1,"iter"], warmup=bed[1,"nWarmups"], bugs.file = file.path(workdir, paste0(bed[,"sequential_number"], "_jags"), "01b_multilevel_intonly_model.bugs"), inits="auto", time.units="secs" )
        res.ja<- data.frame ( bed, software = "jags", runtime.translate = NA, runtime.compile = NA, runtime.samplingfunction = NA, t(fit[["runtimes"]]), stringsAsFactors = FALSE)
        colnames(fit[["samples"]]) <- car::recode(paste0("c", colnames(fit[["samples"]])), "'cparameter'='prm'")
        res.ja<- data.frame ( rbind(res.ja,res.ja,res.ja),fit[["samples"]], stringsAsFactors = FALSE)
     # second: jags (use identical data). adapt priors in bugs file
        if(bed[,"prior"] != 0.001) {
           rows <- grep("inv_gamma", stan)
           stan[rows] <- gsub("0.001", unique(bed[,"prior"]), stan[rows])
           rows2<- grep("normal\\( 0", stan)
           stan[rows2] <- gsub("10000", unique(car::recode(bed[,"prior"],"0.01 = 2500; 0.1 = 100; 1 = 25")), stan[rows2])
        }
        dir.create(file.path(workdir, unique(paste0(bed[,"sequential_number"], "_stan"))))
        write(stan,file.path(workdir, unique(paste0(bed[,"sequential_number"], "_stan")), "02b_multilevel_intonly_model.stan"), sep="\n")
        dlist <- list(J = ncol(d), P = nrow(d), S=S, ybar = as.vector(y.bar))   ### stan does not allow points in object names
        begin <- Sys.time()                                                     ### first step: translate
        transl<- stanc(file = file.path(workdir, unique(paste0(bed[,"sequential_number"], "_stan")), "02b_multilevel_intonly_model.stan"), isystem = file.path(workdir, unique(paste0(bed[,"sequential_number"], "_stan"))))
        t1    <- Sys.time() - begin
        units(t1) <- "secs"
        begin2<- Sys.time()                                                     ### second step: compile
        comp  <- stan_model(stanc_ret = transl, auto_write = TRUE, isystem = file.path(workdir, unique(paste0(bed[,"sequential_number"], "_stan"))))
        t2    <- Sys.time() - begin2
        units(t2) <- "secs"
        begin3<- Sys.time()                                                     ### third step: sampling
        fit   <- sampling(object = comp, data = dlist, iter = bed[1,"iter"] + bed[1,"nWarmups"], chains = 1, warmup = bed[1,"nWarmups"], init = list( list(mu=0, prectheta = 1, preceps = 1)), cores = 1)
        t3    <- Sys.time() - begin3
        units(t3) <- "secs"
        tmes  <- attributes(fit@sim$samples[[1]])[["elapsed_time"]]
        res.st<- data.frame ( bed, software="stan", runtime.translate = as.numeric(t1), runtime.compile = as.numeric(t2), runtime.samplingfunction = as.numeric(t3), runtime.warmup = tmes[["warmup"]], runtime.run = tmes[["sample"]], runtime = tmes[["warmup"]]+tmes[["sample"]], stringsAsFactors = FALSE)
        chains<- as.data.frame(fit@sim[["samples"]][[1]])
        chainSelect      <- as.data.frame(t(chains[(bed[1,"nWarmups"]+1):nrow(chains),1:3]))
        colnames(chainSelect) <- paste0("c", as.character(1:bed[1,"iter"]))
        res.st<- data.frame ( rbind(res.st,res.st,res.st), chainSelect, stringsAsFactors = FALSE)
        res.st[,"prm"]      <- car::recode(rownames(chainSelect), "'sigma2theta'='sigma2.theta'; 'sigma2eps'='sigma2.eps'")
        resAll<- rbind(res.ja, res.st)
        save(resAll, file = file.path(outdir, paste0("results_",bed[,"sequential_number"])))
}
)
